@@ -9,14 +9,14 @@ import os
 import socket
 from subprocess import PIPE, Popen
 
-from xpra.scripts.main import InitException, InitExit, shellquote
+from xpra.scripts.main import InitException, InitExit, shellquote, host_target_string
 from xpra.platform.paths import get_xpra_command, get_ssh_known_hosts_files
 from xpra.platform import get_username
 from xpra.scripts.config import TRUE_OPTIONS
 from xpra.net.bytestreams import SocketConnection, SOCKET_TIMEOUT, ConnectionClosedException
 from xpra.exit_codes import EXIT_SSH_KEY_FAILURE, EXIT_SSH_FAILURE
 from xpra.os_util import (
-    bytestostr, osexpand, monotonic_time,
+    bytestostr, osexpand, monotonic_time, load_binary_file,
     setsid, nomodule_context, umask_context,
     is_WSL, WIN32, OSX, POSIX,
     )
@@ -44,22 +44,6 @@ PASSWORD_RETRY = envint("XPRA_SSH_PASSWORD_RETRY", 2)
 assert PASSWORD_RETRY>=0
 
 
-def ssh_target_string(display_desc):
-    target = "ssh://"
-    username = display_desc.get("username")
-    if username:
-        target += "%s@" % username
-    host = display_desc.get("host")
-    target += host
-    ssh_port = display_desc.get("ssh-port")
-    if ssh_port:
-        target += ":%i" % ssh_port
-    display = display_desc.get("display")
-    if display and display.startswith(":"):
-        display = display[1:]
-    target += "/%s" % (display or "")
-    return target
-
 
 def keymd5(k):
     import binascii
@@ -83,17 +67,37 @@ def exec_dialog_subprocess(cmd):
             env = os.environ.copy()
             env["XPRA_LOG_TO_FILE"] = "0"
             kwargs["env"] = env
-        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, **kwargs)
+        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, **kwargs)
+        stdout = []
+        stderr = []
+        from xpra.gtk_common.gobject_compat import import_gtk
+        gtk = import_gtk()
+        def read_thread(fd, out):
+            while proc.poll() is None:
+                try:
+                    v = fd.read()
+                    if v:
+                        out.append(v)
+                except:
+                    time.sleep(0.1)
+            try:
+                gtk.main_quit()
+            except:
+                pass
+        from xpra.make_thread import start_thread
+        start_thread(read_thread, "dialog-stdout-reader", True, (proc.stdout, stdout))
+        start_thread(read_thread, "dialog-stderr-reader", True, (proc.stderr, stderr))
         if is_WSL():
-            #WSL needs to wait before calling communicate?
+            #WSL needs to wait before calling communicate,
+            #is this still needed now that we read using threads?
             proc.wait()
-        stdout, stderr = proc.communicate()
-        log("exec_dialog_subprocess(%s)", cmd)
+        gtk.main()
+        log("exec_dialog_subprocess(%s) returncode=%s", cmd, proc.poll())
         if stderr:
             log.warn("Warning: dialog process error output:")
-            for x in stderr.splitlines():
+            for x in (b"".join(stderr)).decode().splitlines():
                 log.warn(" %s", x)
-        return proc.returncode, stdout
+        return proc.returncode, (b"".join(stdout)).decode()
     except Exception as e:
         log("exec_dialog_subprocess(..)", exc_info=True)
         log.error("Error: failed to execute the dialog subcommand")
@@ -109,6 +113,7 @@ def dialog_confirm(title, prompt, qinfo="", icon="", buttons=(("OK", 1),)):
     for label, code in buttons:
         cmd.append(nonl(label))
         cmd.append(str(code))
+    log("dialog_confirm%s", (title, prompt, qinfo, icon, buttons))
     return exec_dialog_subprocess(cmd)
 
 def confirm_key(info=()):
@@ -124,6 +129,7 @@ def confirm_key(info=()):
         r = code==200
         log.info("host key %sconfirmed", ["not ", ""][r])
         return r
+    log("confirm_key(%s) will use stdin prompt", nonl(info))
     prompt = "Are you sure you want to continue connecting (yes/NO)? "
     sys.stderr.write(os.linesep.join(info)+os.linesep+prompt)
     v = sys.stdin.readline().rstrip(os.linesep)
@@ -147,7 +153,7 @@ def input_pass(prompt):
 
 class SSHSocketConnection(SocketConnection):
 
-    def __init__(self, ssh_channel, sock, sockname, peername, target, info={}):
+    def __init__(self, ssh_channel, sock, sockname, peername, target, info=None):
         SocketConnection.__init__(self, ssh_channel, sockname, peername, target, "ssh", info)
         self._raw_socket = sock
 
@@ -195,9 +201,29 @@ class SSHSocketConnection(SocketConnection):
 class SSHProxyCommandConnection(SSHSocketConnection):
     def __init__(self, ssh_channel, peername, target, info):
         SSHSocketConnection.__init__(self, ssh_channel, None, None, peername, target, info)
+        self.process = None
+
+    def error_is_closed(self, e):
+        p = self.process
+        if p:
+            #if the process has terminated,
+            #then the connection must be closed:
+            if p[0].poll() is not None:
+                return True
+        return SSHSocketConnection.error_is_closed(self, e)
 
     def get_socket_info(self):
-        return {}
+        p = self.process
+        if not p:
+            return {}
+        proc, _ssh, cmd = p
+        return {
+            "process" : {
+                "pid"       : proc.pid,
+                "returncode": proc.returncode,
+                "command"   : cmd,
+                }
+            }
 
 
 def ssh_paramiko_connect_to(display_desc):
@@ -210,11 +236,12 @@ def ssh_paramiko_connect_to(display_desc):
     if "proxy_host" in display_desc:
         display_desc.setdefault("proxy_username", get_username())
     password = display_desc.get("password")
-    target = ssh_target_string(display_desc)
     remote_xpra = display_desc["remote_xpra"]
     proxy_command = display_desc["proxy_command"]       #ie: "_proxy_start"
     socket_dir = display_desc.get("socket_dir")
+    display = display_desc.get("display")
     display_as_args = display_desc["display_as_args"]   #ie: "--start=xterm :10"
+    keyfiles = None
     socket_info = {
             "host"  : host,
             "port"  : port,
@@ -231,9 +258,14 @@ def ssh_paramiko_connect_to(display_desc):
             host_config = ssh_config.lookup(host)
             if host_config:
                 host = host_config.get("hostname", host)
-                username = host_config.get("username", username)
+                username = host_config.get("user", username)
                 port = host_config.get("port", port)
+                try:
+                    port = int(port)
+                except ValueError:
+                    raise InitException("invalid port specified: '%s'" % port)
                 proxycommand = host_config.get("proxycommand")
+                keyfiles = host_config.get("identityfile")
                 if proxycommand:
                     sock = ProxyCommand(proxycommand)
                     from xpra.child_reaper import getChildReaper
@@ -247,10 +279,12 @@ def ssh_paramiko_connect_to(display_desc):
                     transport = ssh_client.get_transport()
                     do_ssh_paramiko_connect_to(transport, host,
                                                username, password,
-                                               host_config or ssh_config.lookup("*"))
+                                               host_config or ssh_config.lookup("*"),
+                                               keyfiles)
                     chan = paramiko_run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
                     peername = (host, port)
-                    conn = SSHProxyCommandConnection(chan, peername, target, socket_info)
+                    conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
+                    conn.target = host_target_string("ssh", username, host, port, display)
                     conn.timeout = SOCKET_TIMEOUT
                     conn.start_stderr_reader()
                     conn.process = (sock.process, "ssh", cmd)
@@ -277,7 +311,8 @@ def ssh_paramiko_connect_to(display_desc):
             proxy_host_config = ssh_config.lookup(host)
             do_ssh_paramiko_connect_to(middle_transport, proxy_host,
                                        proxy_username, proxy_password,
-                                       proxy_host_config or ssh_config.lookup("*"))
+                                       proxy_host_config or ssh_config.lookup("*"),
+                                       keyfiles)
             log("Opening proxy channel")
             chan_to_middle = middle_transport.open_channel("direct-tcpip", (host, port), ('localhost', 0))
 
@@ -288,11 +323,18 @@ def ssh_paramiko_connect_to(display_desc):
             except SSHException as e:
                 log("start_client()", exc_info=True)
                 raise InitException("SSH negotiation failed: %s" % e)
-            do_ssh_paramiko_connect_to(transport, host, username, password, host_config or ssh_config.lookup("*"))
+            do_ssh_paramiko_connect_to(transport, host,
+                                       username, password,
+                                       host_config or ssh_config.lookup("*"),
+                                       keyfiles)
             chan = paramiko_run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
 
             peername = (host, port)
-            conn = SSHProxyCommandConnection(chan, peername, target, socket_info)
+            conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
+            conn.target = "%s via %s" % (
+                host_target_string("ssh", username, host, port, display),
+                host_target_string("ssh", proxy_username, proxy_host, proxy_port, None),
+                )
             conn.timeout = SOCKET_TIMEOUT
             conn.start_stderr_reader()
             return conn
@@ -310,9 +352,12 @@ def ssh_paramiko_connect_to(display_desc):
         except SSHException as e:
             log("start_client()", exc_info=True)
             raise InitException("SSH negotiation failed: %s" % e)
-        do_ssh_paramiko_connect_to(transport, host, username, password, host_config or ssh_config.lookup("*"))
+        do_ssh_paramiko_connect_to(transport, host, username, password,
+                                   host_config or ssh_config.lookup("*"),
+                                   keyfiles)
         chan = paramiko_run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
-        conn = SSHSocketConnection(chan, sock, sockname, peername, target, socket_info)
+        conn = SSHSocketConnection(chan, sock, sockname, peername, (host, port), socket_info)
+        conn.target = host_target_string("ssh", username, host, port, display)
         conn.timeout = SOCKET_TIMEOUT
         conn.start_stderr_reader()
         return conn
@@ -325,12 +370,14 @@ class nogssapi_context(nomodule_context):
         nomodule_context.__init__(self, "gssapi")
 
 
-def do_ssh_paramiko_connect_to(transport, host, username, password, host_config=None):
-    from paramiko import SSHException, RSAKey, PasswordRequiredException
+def do_ssh_paramiko_connect_to(transport, host, username, password, host_config=None, keyfiles=None):
+    from paramiko import SSHException, PasswordRequiredException
     from paramiko.agent import Agent
     from paramiko.hostkeys import HostKeys
-    log("do_ssh_paramiko_connect_to%s", (transport, host, username, password, host_config))
+    log("do_ssh_paramiko_connect_to%s", (transport, host, username, password, host_config, keyfiles))
     log("SSH transport %s", transport)
+    if not keyfiles:
+        keyfiles = [osexpand(os.path.join("~/", ".ssh", keyfile)) for keyfile in ("id_rsa", "id_dsa")]
 
     host_key = transport.get_remote_server_key()
     assert host_key, "no remote server key"
@@ -371,7 +418,7 @@ def do_ssh_paramiko_connect_to(transport, host, username, password, host_config=
                         log("verifyhostkeydns failed", exc_info=True)
                         log.warn("Warning: cannot check SSHFP DNS records")
                         log.warn(" %s", e)
-            log.info("dnscheck=%s", dnscheck)
+            log("dnscheck=%s", dnscheck)
             def adddnscheckinfo(q):
                 if dnscheck is not True:
                     if dnscheck:
@@ -477,27 +524,47 @@ keymd5(host_key),
                 log.info("agent authentication failed, tried %i key%s", len(agent_keys), engs(agent_keys))
 
     def auth_publickey():
-        log("trying public key authentication")
-        for keyfile in ("id_rsa", "id_dsa"):
-            keyfile_path = osexpand(os.path.join("~/", ".ssh", keyfile))
+        log("trying public key authentication using %s", keyfiles)
+        for keyfile_path in keyfiles:
             if not os.path.exists(keyfile_path):
                 log("no keyfile at '%s'", keyfile_path)
                 continue
+            log("trying '%s'", keyfile_path)
+            key_data = load_binary_file(keyfile_path)
+            if key_data and key_data.find(b"BEGIN OPENSSH PRIVATE KEY")>=0:
+                log.warn("Warning: private key '%s'", keyfile_path)
+                log.warn(" this file seems to be using OpenSSH's own format")
+                log.warn(" please convert it to something more standard (ie: PEM)")
+                log.warn(" so it can be used with the paramiko backend")
+                log.warn(" or switch to the OpenSSH backend with '--ssh=ssh'")
             key = None
-            try:
-                key = RSAKey.from_private_key_file(keyfile_path)
-            except PasswordRequiredException:
-                log("%s keyfile requires a passphrase", keyfile_path)
-                passphrase = input_pass("please enter the passphrase for %s:" % (keyfile_path,))
-                if passphrase:
-                    try:
-                        key = RSAKey.from_private_key_file(keyfile_path, passphrase)
-                    except SSHException as e:
-                        log("from_private_key_file", exc_info=True)
-                        log.info("cannot load key from file '%s':", keyfile_path)
-                        log.info(" %s", e)
+            import paramiko
+            for pkey_classname in ("RSA", "DSS", "ECDSA", "Ed25519"):
+                pkey_class = getattr(paramiko, "%sKey" % pkey_classname, None)
+                if pkey_class is None:
+                    log("no %s key type", pkey_classname)
+                    continue
+                log("trying to load as %s", pkey_classname)
+                try:
+                    key = pkey_class.from_private_key_file(keyfile_path)
+                    log.info("loaded %s private key from '%s'", pkey_classname, keyfile_path)
+                    break
+                except PasswordRequiredException as e:
+                    log("%s keyfile requires a passphrase; %s", keyfile_path, e)
+                    passphrase = input_pass("please enter the passphrase for %s:" % (keyfile_path,))
+                    if passphrase:
+                        try:
+                            key = pkey_class.from_private_key_file(keyfile_path, passphrase)
+                            log.info("loaded %s private key from '%s'", pkey_classname, keyfile_path)
+                        except SSHException as e:
+                            log("from_private_key_file", exc_info=True)
+                            log.info("cannot load key from file '%s':", keyfile_path)
+                            log.info(" %s", e)
+                    break
+                except Exception as e:
+                    log("auth_publickey() loading as %s", pkey_classname, exc_info=True)
             if key:
-                log("auth_publickey using %s: %s", keyfile_path, keymd5(key))
+                log("auth_publickey using %s as %s: %s", keyfile_path, pkey_classname, keymd5(key))
                 try:
                     transport.auth_publickey(username, key)
                 except SSHException as e:
@@ -507,6 +574,8 @@ keymd5(host_key),
                 else:
                     if transport.is_authenticated():
                         break
+            else:
+                log.error("Error: cannot load private key '%s'", keyfile_path)
 
     def auth_none():
         log("trying none authentication")
@@ -522,7 +591,7 @@ keymd5(host_key),
         except SSHException as e:
             log("auth_password(..)", exc_info=True)
             log.info("SSH password authentication failed:")
-            log.info(" %s", e.message)
+            log.info(" %s", getattr(e, "message", e))
 
     def auth_interactive():
         log("trying interactive authentication")
@@ -544,7 +613,7 @@ keymd5(host_key),
         except SSHException as e:
             log("auth_interactive(..)", exc_info=True)
             log.info("SSH password authentication failed:")
-            log.info(" %s", e.message)
+            log.info(" %s", getattr(e, "message", e))
 
     banner = transport.get_banner()
     if banner:
@@ -558,6 +627,9 @@ keymd5(host_key),
     if not transport.is_authenticated() and NONE_AUTH:
         auth_none()
 
+    if not transport.is_authenticated() and AGENT_AUTH:
+        auth_agent()
+
     # Some people do two-factor using KEY_AUTH to kick things off, so this happens first
     if not transport.is_authenticated() and KEY_AUTH:
         auth_publickey()
@@ -568,9 +640,6 @@ keymd5(host_key),
     if not transport.is_authenticated() and PASSWORD_AUTH and password:
         auth_password()
 
-    if not transport.is_authenticated() and AGENT_AUTH:
-        auth_agent()
-
     if not transport.is_authenticated() and PASSWORD_AUTH and not password:
         for _ in range(1+PASSWORD_RETRY):
             password = input_pass("please enter the SSH password for %s@%s:" % (username, host))
@@ -785,15 +854,19 @@ def ssh_exec_connect_to(display_desc, opts=None, debug_cb=None, ssh_fail_cb=ssh_
                 child.terminate()
         except Exception as e:
             print("error trying to stop ssh tunnel process: %s" % e)
+    host = display_desc["host"]
+    port = display_desc.get("ssh-port", 22)
+    username = display_desc.get("username")
+    display = display_desc.get("display")
     info = {
-        "host"  : display_desc["host"],
-        "port"  : display_desc.get("ssh-port", 22),
+        "host"  : host,
+        "port"  : port,
         }
     from xpra.net.bytestreams import TwoFileConnection
-    target = ssh_target_string(display_desc)
     conn = TwoFileConnection(child.stdin, child.stdout,
-                             abort_test, target=target,
+                             abort_test, target=(host, port),
                              socktype="ssh", close_cb=stop_tunnel, info=info)
+    conn.endpoint = host_target_string("ssh", username, host, port, display)
     conn.timeout = 0            #taken care of by abort_test
     conn.process = (child, "ssh", cmd)
     return conn
