@@ -828,8 +828,18 @@ class XpraServer(gobject.GObject, X11ServerBase):
         assert len(event.data)==5
         #x_root, y_root, direction, button, source_indication = event.data
         wid = self._window_to_id[window]
-        for ss in self._server_sources.values():
-            ss.initiate_moveresize(wid, window, *event.data)
+        #find clients that handle windows:
+        from xpra.server.source.windows_mixin import WindowsMixin
+        wsources = [ss for ss in self._server_sources.values() if isinstance(ss, WindowsMixin)]
+        if not wsources:
+            return
+        #prefer the "UI driver" if we find it:
+        driversources = [ss for ss in wsources if self.ui_driver==ss.uuid]
+        if driversources:
+            driversources[0].initiate_moveresize(wid, window, *event.data)
+            return
+        #otherwise, fallback to the first one:
+        wsources[0].initiate_moveresize(wid, window, *event.data)
 
 
     def _raised_window(self, window, event):
@@ -992,6 +1002,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 pwid = packet[10]
                 pointer = packet[11]
                 modifiers = packet[12]
+                if pwid==wid and window.is_OR():
+                    #some clients may send the OR window wid
+                    #this causes focus issues (see #1999)
+                    pwid = -1
+                mouselog("configure pointer data: %s", (pwid, pointer, modifiers))
                 if self._process_mouse_common(proto, pwid, pointer):
                     #only update modifiers if the window is in focus:
                     if self._has_focus==wid:
